import unittest

import numpy as np
import pytest

from laserchicken import keys
from laserchicken.feature_extractor.band_ratio_feature_extractor import BandRatioFeatureExtractor
from laserchicken.test_tools import create_point_cloud
from laserchicken.volume_specification import Cell, Sphere, Cube, InfiniteCylinder

UPPER_LIMIT = 5
LOWER_LIMIT = 3

VALUE_BELOW_LIMIT = -1
VALUE_WITHIN_LIMITS = 4
VALUE_ABOVE_LIMITS = 11


class TestBandRatioFeatureExtractorSimpleArtificialData(unittest.TestCase):
    """Test echo ratio extractor on artificial data."""

    def test_zero_points_nan_outcome(self):
        """Zero points should return NaN."""
        point_cloud = create_point_cloud(np.zeros(0), np.zeros(0), np.zeros(0))
        targets = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))
        result = BandRatioFeatureExtractor(2, 4).extract(point_cloud, [[]], targets, 0, Cell(4))
        self.assertTrue(np.isnan(result[0]))

    def test_no_upper_bound_correct_outcome(self):
        """No upper bound should return all points above."""
        n = 10
        point_cloud = create_point_cloud(np.zeros(n), np.zeros(n), np.hstack([np.zeros(8), np.ones(2)]))
        targets = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))
        result = BandRatioFeatureExtractor(0.5, None).extract(point_cloud, [range(n)], targets, 0, Cell(4))
        np.testing.assert_equal(result[0], 0.2)

    def test_no_lower_bound_correct_outcome(self):
        """No lower bound should return all points below."""
        n = 10
        point_cloud = create_point_cloud(np.zeros(n), np.zeros(n), np.hstack([np.zeros(8), np.ones(2)]))
        targets = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))
        result = BandRatioFeatureExtractor(None, 0.5).extract(point_cloud, [range(n)], targets, 0, Cell(4))
        np.testing.assert_equal(result[0], 0.8)

    def test_use_optional_data_key(self):
        """Should use data under the given data key (normalized z)."""
        n = 10
        zeros = np.zeros(n)
        point_cloud = create_point_cloud(zeros, zeros, zeros, normalized_z=np.hstack([np.zeros(8), np.ones(2)]))
        targets = create_point_cloud(np.zeros(1), np.zeros(1), np.zeros(1))
        extractor = BandRatioFeatureExtractor(None, 0.5, data_key=keys.normalized_height)

        result = extractor.extract(point_cloud, [range(n)], targets, 0, Cell(4))

        np.testing.assert_equal(result[0], 0.8)

    def test_simple_case_correct(self):
        """Ratio of points within with respect to total should be returned."""
        n_below_limit = 10
        n_within = 40
        n_above_limit = 50
        expected_ratio = 0.4  # which is n_within/(n_below_limit+n_within+n_above_limit)

        assert_expected_ratio(expected_ratio, n_below_limit, n_within, n_above_limit)

    def test_mixed_neighborhoods(self):
        """Test with two neighborhoods of different lengths (in parallel)."""
        expected_ratios = [0.2, 0.6]
        n_below_limits = [20, 0]
        n_withins = [10, 60]
        n_above_limits = [20, 40]
        assert_expected_ratios(expected_ratios, n_below_limits, n_withins, n_above_limits)

    def test_sphere_volume_raise(self):
        with pytest.raises(ValueError):
            assert_expected_ratio(volume=Sphere(5))

    def test_cube_volume_raise(self):
        with pytest.raises(ValueError):
            assert_expected_ratio(volume=Cube(5))

    def test_cell_volume(self):
        assert_expected_ratio(volume=Cell(5))

    def test_cell_volume(self):
        assert_expected_ratio(volume=InfiniteCylinder(5))

    def test_provides_simple(self):
        self.assertEqual(['band_ratio_6<z<20'], BandRatioFeatureExtractor(6, 20).provides())

    def test_provides_with_only_upper_limit(self):
        self.assertEqual(['band_ratio_z<20'], BandRatioFeatureExtractor(None, 20).provides())

    def test_provides_with_only_lower_limit(self):
        self.assertEqual(['band_ratio_20<z'], BandRatioFeatureExtractor(20, None).provides())

    def test_provides_with_zero_lower_limit(self):
        self.assertEqual(['band_ratio_0<z'], BandRatioFeatureExtractor(0, None).provides())

    def test_provides_with_zero_upper_limit(self):
        self.assertEqual(['band_ratio_z<0'], BandRatioFeatureExtractor(None, 0).provides())

    def test_provides_with_data_key(self):
        self.assertEqual(['band_ratio_1<normalized_height<3'],
                         BandRatioFeatureExtractor(1, 3, data_key=keys.normalized_height).provides())


def assert_expected_ratio(expected_ratio=0.4, n_below_limit=10, n_within=40, n_above_limit=50, volume=Cell(4)):
    assert_expected_ratios(np.array([expected_ratio]),
                           np.array([n_below_limit]),
                           np.array([n_within]),
                           np.array([n_above_limit]),
                           volume)


def assert_expected_ratios(expected_ratios, n_below_limits, n_withins, n_above_limits, volume=Cell(4)):
    n_ratios = len(expected_ratios)
    neighborhoods, point_cloud = generate_test_point_cloud_and_neighborhoods(n_below_limits,
                                                                             n_withins,
                                                                             n_above_limits,
                                                                             n_ratios)
    targets = create_point_cloud(np.zeros(n_ratios), np.zeros(n_ratios), np.zeros(n_ratios))
    extractor = BandRatioFeatureExtractor(LOWER_LIMIT, UPPER_LIMIT)
    result = extractor.extract(point_cloud, neighborhoods, targets, 0, volume)
    np.testing.assert_allclose(result, expected_ratios)


def generate_test_point_cloud_and_neighborhoods(n_below_limits, n_withins, n_above_limits, n_ratios):
    neighborhoods = []
    cursor = 0
    xs = []
    ys = []
    zs = []
    for i in range(n_ratios):
        x_i, y_i, z_i = generate_test_points(n_below_limits[i], n_withins[i], n_above_limits[i])
        xs.append(x_i)
        ys.append(y_i)
        zs.append(z_i)

        n_neighbors = len(x_i)
        neighborhood = list(range(cursor, cursor + n_neighbors))
        neighborhoods += [neighborhood]
        cursor += n_neighbors

    x = np.hstack(xs)
    y = np.hstack(ys)
    z = np.hstack(zs)
    point_cloud = create_point_cloud(x, y, z)
    return neighborhoods, point_cloud


def generate_test_points(n_below_limit, n_within, n_above_limit):
    """Generate a bunch of daata points on the z axis with the specified number"""
    n_neighbors = n_below_limit + n_within + n_above_limit
    x = np.zeros(n_neighbors)
    y = x.copy()
    z = x.copy()
    z[0:n_below_limit] = VALUE_BELOW_LIMIT
    z[n_below_limit:n_within + n_below_limit] = VALUE_WITHIN_LIMITS
    z[n_within + n_below_limit:n_below_limit + n_within + n_above_limit] = VALUE_ABOVE_LIMITS
    return x, y, z



import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.utils import get_xyz_per_neighborhood


class EigenValueVectorizeFeatureExtractor(FeatureExtractor):

    @classmethod
    def requires(cls):
        return []

    @classmethod
    def provides(cls):
        return ['eigenv_1', 'eigenv_2', 'eigenv_3', 'normal_vector_1', 'normal_vector_2', 'normal_vector_3', 'slope']

    def extract(self, sourcepc, neighborhoods, targetpc, targetindex, volume):
        if not (isinstance(neighborhoods[0], list) or isinstance(neighborhoods[0], range)):
            neighborhoods = [neighborhoods]

        xyz_grp = get_xyz_per_neighborhood(sourcepc, neighborhoods)
        self._mask_rows_with_too_few_points(xyz_grp)

        e_vals, eigvects = self._get_eigen_values_and_vectors(xyz_grp)
        normals = self._get_normals(eigvects)
        alpha = np.arccos(np.dot(normals, np.array([0., 0., 1.])))
        slope = np.tan(alpha)

        return e_vals[:, 0], e_vals[:, 1], e_vals[:, 2], normals[:, 0], normals[:, 1], normals[:, 2], slope

    @staticmethod
    def _get_cov(xyz):
        n_max = xyz.shape[2]
        n = (xyz.mask.sum(axis=2, keepdims=True) * -1) + n_max
        m = xyz - xyz.sum(2, keepdims=True) / n
        return np.einsum('ijk,ilk->ijl', m, m) / (n - 1)

    def _get_eigen_values_and_vectors(self, xyz_grp):
        cov_mat = self._get_cov(xyz_grp)
        eigval, eigvects = np.linalg.eig(cov_mat)
        return self._sort_eigen_values_and_vectors_by_value(eigval, eigvects)

    def _sort_eigen_values_and_vectors_by_value(self, eigval, eigvects):
        val_indices = np.argsort(eigval, axis=1)[:, ::-1]
        ordered_vects = self._reorder_vectors(eigvects, val_indices)
        ordered_eigval = np.take_along_axis(eigval, val_indices, axis=1)
        return ordered_eigval, ordered_vects

    @staticmethod
    def _reorder_vectors(eigen_vectors, new_vector_indices):
        """
        Swaps the order of the eigen vectors according to the given new indices. Note that each single vector is kept
        unchanged.
        :param eigen_vectors:
        :param new_vector_indices:
        :return: The input eigen vectors in their new order.
        """
        # I'm reordering vectors in the last two dimensions while keeping the first dimension unchanged. I can't find
        # any existing functions that do this exactly the way it needs to be sorted. Therefore I have to transpose
        # and reshape and even repeat some indices to get the behavior I'm looking for.
        vects_t = eigen_vectors.transpose([0, 2, 1])  # Eigen vectors used as column vectors. Making row vectors here.
        flattened_vects_t = vects_t.reshape(-1, 3 * 3)  # Flatten the (eigen) vector dimension
        vect_indices = np.zeros_like(flattened_vects_t, dtype=np.int) + [0, 1, 2, 0, 1, 2, 0, 1, 2]  # 0,1,2 for x,y,z
        vect_indices[:, :3] += new_vector_indices[:, 0:1] * 3  # Because x,y,z, indices have to be increased by 3.
        vect_indices[:, 3:6] += new_vector_indices[:, 1:2] * 3
        vect_indices[:, 6:9] += new_vector_indices[:, 2:3] * 3
        ordered_flattened_vects_t = np.take_along_axis(flattened_vects_t, vect_indices, axis=1)
        ordered_vects_t = ordered_flattened_vects_t.reshape(eigen_vectors.shape)
        ordered_vects = ordered_vects_t
        return ordered_vects.transpose([0, 2, 1])

    @staticmethod
    def _get_normals(eigen_vectors):
        """
        Return normals given eigen vectors. For all instances, all elements of the 3th vector (= normal vector) are
        taken. Vectors that are pointing downward are flipped so only upward vectors are returned. :param eigvects:
        :return:
        """
        normals = eigen_vectors[:, :, 2]
        mask_of_downward_pointing_vectors = normals[:, 2] < 0
        normals[mask_of_downward_pointing_vectors] = -normals[mask_of_downward_pointing_vectors]
        return normals

    @staticmethod
    def _mask_rows_with_too_few_points(xyz_grp):
        minimum_for_calculation = 3
        invalid_cells = xyz_grp.mask == False
        invalid_neighbors = np.any(invalid_cells, axis=1)
        invalid_rows = np.sum(invalid_neighbors, axis=1) < minimum_for_calculation
        xyz_grp.mask[invalid_rows, :, :] = True



import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point


class VarianceFeatureExtractor(FeatureExtractor):
    """Calculates the variation on the z axis."""
    def __init__(self, data_key='z'):
        self.data_key = data_key

    @classmethod
    def requires(cls):
        return []

    def provides(self):
        return ['var_' + self.data_key]

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_indices, volume_description):
        return [self._extract_one(point_cloud, neighborhood) for neighborhood in neighborhoods]

    def _extract_one(self, source_point_cloud, neighborhood):
        if neighborhood:
            source_data = source_point_cloud[point][self.data_key]['data'][neighborhood]
            var_z = np.var(source_data)
        else:
            var_z = np.NaN
        return var_z



# Name of point data section in point cloud structure
point = 'vertex'

# Name of the normalized height point attribute
normalized_height = 'normalized_height'

# Name of the intensity point attribute
intensity = 'intensity'

#
point_cloud = 'pointcloud'

#
provenance = 'log'



"""Functions that select points from a point cloud and put them in a new point cloud."""

import math
import os
import shapefile
import shapely
import sys
from shapely.geometry import Point
from shapely.errors import WKTReadingError
from shapely.wkt import loads
from shapely.geometry import box
import numpy as np

from laserchicken.keys import point
from laserchicken import kd_tree
from laserchicken.utils import copy_point_cloud, add_metadata


def select_equal(point_cloud, attribute, value, return_mask=False):
    """
    Return the selection of the input point cloud that contains only points with a given attribute equal to some value.
    If a list of values is given, select the points corresponding to any of the provided values.

    :param point_cloud: Input point cloud.
    :param attribute: The attribute name used for selection
    :param value: The value(s) to compare the attribute to
    :param return_mask: If true, return the mask corresponding to the selection
    :return:
    """
    _check_valid_arguments(attribute, point_cloud)
    # broadcast using shape of the values
    mask = point_cloud[point][attribute]['data'] == np.array(value)[..., None]
    if mask.ndim > 1:
        mask = np.any(mask, axis=0)  # reduce
    if return_mask:
        return mask
    point_cloud_filtered = copy_point_cloud(point_cloud, mask)
    add_metadata(point_cloud_filtered, sys.modules[__name__],
                 {'attribute': attribute, 'value': value})
    return point_cloud_filtered


def select_above(point_cloud, attribute, threshold, return_mask=False):
    """
    Return the selection of the input point cloud that contains only points with a given attribute above some value.

    :param point_cloud: Input point cloud
    :param attribute: The attribute name used for selection
    :param threshold: The threshold value used for selection
    :param return_mask: If true, return the mask corresponding to the selection
    :return:
    """
    _check_valid_arguments(attribute, point_cloud)
    mask = point_cloud[point][attribute]['data'] > threshold
    if return_mask:
        return mask
    point_cloud_filtered = copy_point_cloud(point_cloud, mask)
    add_metadata(point_cloud_filtered, sys.modules[__name__],
                 {'attribute': attribute, 'threshold': threshold})
    return point_cloud_filtered


def select_below(point_cloud, attribute, threshold, return_mask=False):
    """
    Return the selection of the input point cloud that contains only points with a given attribute below some value.

    :param point_cloud: Input point cloud
    :param attribute: The attribute name used for selection
    :param threshold: The threshold value used for selection
    :param return_mask: If true, return the mask corresponding to the selection
    :return:
    """
    _check_valid_arguments(attribute, point_cloud)
    mask = point_cloud[point][attribute]['data'] < threshold
    if return_mask:
        return mask
    point_cloud_filtered = copy_point_cloud(point_cloud, mask)
    add_metadata(point_cloud_filtered, sys.modules[__name__],
                 {'attribute': attribute, 'threshold': threshold})
    return point_cloud_filtered


def _check_valid_arguments(attribute, point_cloud):
    """
    Raise if arguments are not valid for select_above/select_below functions.

    :param attribute:
    :param point_cloud:
    :return: None
    """
    if point_cloud is None:
        raise ValueError('Input point cloud cannot be None.')
    if attribute not in point_cloud[point]:
        raise ValueError('Attribute key {} for selection not found in point cloud.'.format(attribute))


def select_polygon(point_cloud, polygon_string, read_from_file=False, return_mask=False):
    """
    Return the selection of the input point cloud that contains only points within a given polygon.

    :param point_cloud: Input point cloud
    :param polygon_string: Polygon, either defined in a WKT string or in a file (WKT and ESRI formats supported)
    :param read_from_file: if true, polygon is expected to be the name of the file where the polygon is defined
    :param return_mask: if true, return a mask of selected points, rather than point cloud
    :return:
    """
    if point_cloud is None:
        raise ValueError('Input point cloud cannot be None.')
    if not isinstance(polygon_string, str):
        raise ValueError('Polygon (or its filename) should be a string')
    if read_from_file:
        format = os.path.splitext(polygon_string)[1].lower()
        reader = _get_polygon_reader(format)
        polygon = reader(polygon_string)
    else:
        polygon = _load_polygon(polygon_string)
    
    if isinstance(polygon, shapely.geometry.polygon.Polygon) and polygon.is_valid:
        points_in = _contains(point_cloud, polygon)
    elif isinstance(polygon,shapely.geometry.multipolygon.MultiPolygon) and polygon.is_valid:
        points_in = []
        count=1
        for poly in polygon:
            if not(count%200) or count==len(polygon):
                print('Checking polygon {}/{}...'.format(count, len(polygon)))
            points_in.extend(_contains(point_cloud, poly))
            count=count+1
        print('{} points found in {} polygons.'.format(len(points_in), len(polygon)))
    else:
        raise ValueError('It is not a Polygon or Multipolygon.')
    
    if return_mask: 
        mask = np.zeros(len(point_cloud['vertex']['x']['data']), dtype=bool)
        mask[points_in] = True
        return mask
    else:
        point_cloud_filtered = copy_point_cloud(point_cloud, points_in)
        add_metadata(point_cloud_filtered, sys.modules[__name__],
                    {'polygon_string': polygon_string,
                    'read_from_file': read_from_file})
        return point_cloud_filtered


def _read_wkt_file(path):
    with open(path) as f:
        content = f.readlines()

    content = [x.strip() for x in content]
    return _load_polygon(content[0])


def _read_shp_file(path):
    shape = shapefile.Reader(path)
    # first feature of the shapefile
    feature = shape.shapeRecords()[0]
    first = feature.shape.__geo_interface__
    # or shp_geom = shape(first) with PyShp)
    shp_geom = shapely.geometry.shape(first)
    return shp_geom


polygon_readers = {
    '.wkt': _read_wkt_file,
    '.shp': _read_shp_file
}


def _get_polygon_reader(format):
    if format not in polygon_readers:
        raise NotImplementedError(
            'Polygon file format {} unknown. Implemented formats are:'.format(format, ','.join(polygon_readers.keys())))
    else:
        return polygon_readers[format]


def _load_polygon(string):
    try:
        return loads(string)
    except WKTReadingError:
        raise ValueError('Polygon is invalid. --> {}'.format(string))


def _contains(pc, polygon):
    """
    Return indices of points in point cloud that are contained by a polygon, i.e., all points within the boundaries of
    Polygon excluding the ones overlaping Polygon's boundaries.
    :param pc: point cloud in
    :param polygon: containing polygon
    :return: point indices
    """
    x = pc[point]['x']['data']
    y = pc[point]['y']['data']
    points_in = []

    mbr = polygon.envelope
    point_box = box(np.min(x), np.min(y), np.max(x), np.max(y))

    if point_box.intersects(mbr):
        (x_min, y_min, x_max, y_max) = mbr.bounds

        rad = math.ceil(math.sqrt(math.pow(x_max - x_min, 2) +
                                  math.pow(y_max - y_min, 2)) / 2)
        p = [x_min + ((x_max - x_min) / 2), y_min + ((y_max - y_min) / 2)]
        tree = kd_tree.get_kdtree_for_pc(pc)
        indices = np.sort(tree.query_ball_point(x=p, r=rad))

        point_id = 0
        for i in indices:            
            if polygon.contains(Point(x[i], y[i])):
                points_in.append(i)
                point_id += 1

    return points_in



import os
import shutil
import unittest

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_equal, assert_almost_equal

from laserchicken import load
from laserchicken.keys import point
from laserchicken.filter import select_below, select_above, select_equal, select_polygon
from laserchicken.test_tools import ComplexTestData


class TestSelectEqual(unittest.TestCase):
    @staticmethod
    def test_selectEqual_None():
        """ None input raises Value Error. """
        assert_none_pc_raises_value_error(select_equal, 'x', 13)

    @staticmethod
    def test_selectEqual_unknownKey():
        """ If key is not in point cloud, raise Value Error. """
        assert_unknown_key_raises_value_error(select_equal)

    @staticmethod
    def test_selectEqual_noneKey():
        """ If key is None, raise Value Error. """
        assert_none_key_raises_value_error(select_equal)

    @staticmethod
    def test_selectEqual_outputIsNotInput():
        """ Select change output, make sure that input hasn't changed. """
        assert_output_is_not_input(select_equal)

    @staticmethod
    def test_selectEqual_outputEmpty():
        """ Select everything equal to something that doesn't exist should result in empty output. """
        pc_in = get_test_data()
        pc_out = select_equal(pc_in, 'return', 5)
        assert_equal(len(pc_out[point]['x']['data']), 0)

    @staticmethod
    def test_selectEqual_outputCorrect():
        """ Correct number of results. """
        pc_in = get_test_data()
        pc_out = select_equal(pc_in, 'return', 1)
        assert_equal(len(pc_out[point]['x']['data']), 2)

    @staticmethod
    def test_selectEqual_multipleValues():
        """ Correct number of results. """
        pc_in = get_test_data()
        pc_out = select_equal(pc_in, 'return', [1, 2])
        assert_equal(len(pc_out[point]['x']['data']), 3)

    @staticmethod
    def test_selectEqual_maskCorrect():
        """ Correct number of results. """
        pc_in = get_test_data()
        mask_out = select_equal(pc_in, 'return', 1, return_mask=True)
        assert_equal(mask_out, np.array([1,1,0], dtype=bool))

    @staticmethod
    def test_selectEqual_maskEmpty():
        """ Correct number of results. """
        pc_in = get_test_data()
        mask_out = select_equal(pc_in, 'return', 3, return_mask=True)
        assert_equal(sum(mask_out), 0)


class TestSelectBelow(unittest.TestCase):
    @staticmethod
    def test_selectBelow_None():
        """ None input raises Value Error. """
        assert_none_pc_raises_value_error(select_below, 'x', 13)

    @staticmethod
    def test_selectBelow_unknownKey():
        """ If key is not in point cloud, raise Value Error. """
        assert_unknown_key_raises_value_error(select_below)

    @staticmethod
    def test_selectBelow_noneKey():
        """ If key is None, raise Value Error. """
        assert_none_key_raises_value_error(select_below)

    @staticmethod
    def test_selectBelow_outputIsNotInput():
        """ Select change output, make sure that input hasn't changed. """
        assert_output_is_not_input(select_below)

    @staticmethod
    def test_selectBelow_everything():
        """ Selecting all point under some super high value should return all the points. """
        pc_in = get_test_data()
        pc_out = select_below_ridiculous_high_z(pc_in)
        assert_equal(len(pc_out[point]['x']['data']), 3)
        assert_points_equal(pc_in, pc_out)

    @staticmethod
    def test_selectBelow_onlyOnePoint():
        """ Selecting below some threshold should only result the correct lowest point. """
        pc_in = get_test_data()
        pc_out = select_below(pc_in, 'z', 3.2)
        assert_equal(len(pc_out[point]['x']['data']), 1)
        assert_almost_equal(pc_out[point]['x']['data'][0], 1.1)
        assert_almost_equal(pc_out[point]['y']['data'][0], 2.1)
        assert_almost_equal(pc_out[point]['z']['data'][0], 3.1)
        assert_almost_equal(pc_out[point]['return']['data'][0], 1)
    
    @staticmethod
    def test_selectBelow_maskCorrect():
        """ Correct number of results. """
        pc_in = get_test_data()
        mask_out = select_below(pc_in, 'return', 2, return_mask=True)
        assert_equal(mask_out, np.array([1,1,0], dtype=bool))

    @staticmethod
    def test_selectBelow_maskAll():
        """ Correct number of results. """
        pc_in = get_test_data()
        mask_out = select_below(pc_in, 'return', 3, return_mask=True)
        assert_equal(sum(mask_out), 3)


class TestSelectAbove(unittest.TestCase):
    @staticmethod
    def test_selectAbove_None():
        """ None input raises Value Error. """
        assert_none_pc_raises_value_error(select_above, 'x', 13)

    @staticmethod
    def test_selectAbove_unknownKey():
        """ If key is not in point cloud, raise Value Error. """
        assert_unknown_key_raises_value_error(select_above)

    @staticmethod
    def test_selectAbove_noneKey():
        """ If key is None, raise Value Error. """
        assert_none_key_raises_value_error(select_above)

    @staticmethod
    def test_selectAbove_outputIsNotInput():
        """ Select change output, make sure that input hasn't changed. """
        assert_output_is_not_input(select_above)

    @staticmethod
    def test_selectAbove_everything():
        """ Selecting all point above some super low value should return all the points. """
        pc_in = get_test_data()
        pc_out = select_above(pc_in, 'z', -100000)
        assert_points_equal(pc_in, pc_out)

    @staticmethod
    def test_selectBelow_onlyOnePoint():
        """ Selecting above some threshold should only result the correct highest point. """
        pc_in = get_test_data()
        pc_out = select_above(pc_in, 'z', 3.2)
        assert_equal(len(pc_out[point]['x']['data']), 1)
        assert_almost_equal(pc_out[point]['x']['data'][0], 1.3)
        assert_almost_equal(pc_out[point]['y']['data'][0], 2.3)
        assert_almost_equal(pc_out[point]['z']['data'][0], 3.3)
        assert_almost_equal(pc_out[point]['return']['data'][0], 2)
    
    @staticmethod
    def test_selectAbove_maskCorrect():
        """ Correct number of results. """
        pc_in = get_test_data()
        mask_out = select_above(pc_in, 'return', 1, return_mask=True)
        assert_equal(mask_out, np.array([0,0,1], dtype=bool))

    @staticmethod
    def test_selectAbove_maskAll():
        """ Correct number of results. """
        pc_in = get_test_data()
        mask_out = select_above(pc_in, 'return', 0, return_mask=True)
        assert_equal(sum(mask_out), 3)


class TestSelectPolygonWKT(unittest.TestCase):
    polygon_around_1_point_ahn2 = ("POLYGON(( 243627.840248 572073.439002, "
                                   "243627.842248 572073.439002, "
                                   "243627.842248 572073.441002, "
                                   "243627.840248 572073.441002, "
                                   "243627.840248 572073.439002))")

    @staticmethod
    def test_points_in_polygon_wkt_None():
        """ None input raises Value Error. """
        assert_none_pc_raises_value_error(select_polygon, "POLYGON((0.,0.,0.))")

    @staticmethod
    def test_points_in_polygon_wkt_noneWKT():
        """ If key is None, raise Value Error. """
        assert_none_wkt_raises_value_error(select_polygon)

    @staticmethod
    def test_points_in_polygon_wkt_Point():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "POINT(6 10)")

    @staticmethod
    def test_points_in_polygon_wkt_Line():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "LINESTRING(3 4,10 50,20 25)")

    @staticmethod
    def test_points_in_polygon_wkt_MultiPoint():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "MULTIPOINT(3.5 5.6, 4.8 10.5)")

    @staticmethod
    def test_points_in_polygon_wkt_MultiLine():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "MULTILINESTRING((3 4,10 50,20 25),(-5 -8,-10 -8,-15 -4))")

    @staticmethod
    def test_points_in_polygon_wkt_MultiPolygon():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in,
                           "MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2)),((3 3,6 2,6 4,3 3)))")

    @staticmethod
    def test_points_in_polygon_wkt_Collection():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "GEOMETRYCOLLECTION(POINT(4 6),LINESTRING(4 6,7 10))")

    @staticmethod
    def test_points_in_polygon_wkt_invalidPolygon():
        """Polygon is not closed so should raise error."""
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in,
                           "POLYGON(( 243590.0 572110.0, 243640.0 572160.0, 243700.0 572110.0, 243640.0 572060.0 ))")

    @staticmethod
    def test_wkt_polygons_contains():
        """ Selecting all points within a Polygon using artificial data. """
        test_data = ComplexTestData()
        pc_in = test_data.get_point_cloud()
        expected_point = np.array(pc_in[point]['x']['data'][0], pc_in[point]['y']['data'][0])
        pc_out = select_polygon(pc_in, test_data.get_wkt_polygon_around_first_point_only())
        assert_equal(len(pc_out[point]['x']['data']), 1)
        selected_point = np.array(pc_out[point]['x']['data'][0], pc_out[point]['y']['data'][0])
        np.testing.assert_allclose(selected_point, expected_point)

    def test_wkt_polygons_contains_single_point(self):
        """Selecting a single point with a tiny polygon. Test for https://github.com/eEcoLiDAR/eEcoLiDAR/issues/64. """
        pc_in = load("testdata/AHN2.las")
        pc_out = select_polygon(pc_in, self.polygon_around_1_point_ahn2)
        x = pc_out[point]['x']['data']
        y = pc_out[point]['y']['data']
        expected_x = 243627.841248
        expected_y = 572073.440002
        assert_equal(len(x), 1)
        assert_almost_equal(x[0], expected_x, 4)
        assert_almost_equal(y[0], expected_y, 4)

    def test_wkt_polygons_contains_original_not_changed(self):
        """Point cloud in should not change by filtering."""
        pc_in = load("testdata/AHN2.las")
        len_x_before = len(pc_in[point]['x']['data'])
        _pc_out = select_polygon(pc_in, self.polygon_around_1_point_ahn2)
        len_x_after = len(pc_in[point]['x']['data'])
        assert_equal(len_x_after, len_x_before)

    @staticmethod
    def test_wkt_polygons_containsEmpty():
        """ Selecting all points within a Polygon. """
        pc_in = load("testdata/AHN2.las")
        pc_out = select_polygon(pc_in,
                                ("POLYGON(( 253590.0 582110.0, "
                                 "253640.0 582160.0, 253700.0 582110.0, "
                                 "253640.0 582060.0, 253590.0 582110.0 ))"))
        x = pc_out[point]['x']['data']
        y = pc_out[point]['y']['data']
        assert (len(x) == 0)
        assert (len(y) == 0)


class TestSelectPolygonWKTFile(unittest.TestCase):
    _test_dir = 'wkt_test_dir'

    @staticmethod
    def test_points_in_polygon_wkt_None():
        """ None input raises Value Error. """
        assert_none_pc_raises_value_error(select_polygon,
                                          "testdata/ahn2_polygon_shp/ahn2_polygon.wkt",
                                          read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_unknownPath():
        """ None input raises Value Error. """
        assert_unknown_path_raises_value_error(select_polygon)

    @staticmethod
    def test_points_in_polygon_wkt_nonePath():
        """ If key is None, raise Value Error. """
        assert_none_path_raises_value_error(select_polygon)

    @staticmethod
    def test_points_in_polygon_wkt_Point():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/point.wkt", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_Line():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/line.wkt", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_MultiPoint():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/multipoint.wkt", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_MultiLine():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/multiline.wkt", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_MultiPolygon():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/multipolygon.wkt", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_Collection():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/collection.wkt", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_wkt_invalidPolygon():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_wkt/invalid_polygon.wkt", read_from_file=True)

    def test_wkt_polygons_contains(self):
        """Selecting all points within a Polygon with artificial data."""
        test_data = ComplexTestData()
        path = os.path.join(self._test_dir, 'wkt_test.wkt')
        with open(path, 'w') as f:
            f.write(test_data.get_wkt_polygon_around_first_point_only())
        pc_in = test_data.get_point_cloud()
        expected_point = np.array(pc_in[point]['x']['data'][0], pc_in[point]['y']['data'][0])

        pc_out = select_polygon(pc_in, path, read_from_file=True)

        assert_equal(len(pc_out[point]['x']['data']), 1)
        selected_point = np.array(pc_out[point]['x']['data'][0], pc_out[point]['y']['data'][0])
        np.testing.assert_allclose(selected_point, expected_point)

    @staticmethod
    def test_wkt_polygons_containsEmpty():
        """ Selecting all points within a Polygon. """
        pc_in = load("testdata/AHN2.las")
        pc_out = select_polygon(pc_in, "testdata/ahn2_geometries_wkt/ahn2_polygon_empty.wkt", read_from_file=True)
        x = pc_out[point]['x']['data']
        y = pc_out[point]['y']['data']
        assert (len(x) == 0)
        assert (len(y) == 0)

    def setUp(self):
        os.mkdir(self._test_dir)

    def tearDown(self):
        shutil.rmtree(self._test_dir)


class TestSelectPolygonSHPFile(unittest.TestCase):
    @staticmethod
    def test_points_in_polygon_shp_None():
        """ None input raises Value Error. """
        assert_none_pc_raises_value_error(select_polygon,
                                          "testdata/ahn2_geometries_shp/point.shp",
                                          read_from_file=True)

    @staticmethod
    def test_points_in_polygon_shp_unknownPath():
        """ None input raises Value Error. """
        assert_unknown_path_raises_value_error(select_polygon)

    @staticmethod
    def test_points_in_polygon_shp_nonePath():
        """ If key is None, raise Value Error. """
        assert_none_path_raises_value_error(select_polygon)

    @staticmethod
    def test_points_in_polygon_shp_Point():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_shp/point.shp", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_shp_Line():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_shp/line.shp", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_shp_MultiPoint():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_shp/multipoint.shp", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_shp_MultiLine():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_shp/multiline.shp", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_shp_Collection():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_shp/collection.shp", read_from_file=True)

    @staticmethod
    def test_points_in_polygon_shp_invalidPolygon():
        pc_in = load("testdata/AHN2.las")
        with pytest.raises(ValueError):
            select_polygon(pc_in, "testdata/ahn2_geometries_shp/invalid_polygon.shp", read_from_file=True)

    @staticmethod
    def test_shp_polygons_contains():
        """ Selecting all points within a Polygon. """
        pc_in = load("testdata/AHN2.las")
        pc_out = select_polygon(pc_in, "testdata/ahn2_geometries_shp/ahn2_polygon.shp", read_from_file=True)
        x = pc_out[point]['x']['data']
        y = pc_out[point]['y']['data']
        # Seemingly redundant 'astype' call: since pandas 0.24 Dataframe() doesn't enforce the given dtype as before
        df_out = pd.DataFrame({'x': x, 'y': y}, dtype=np.int32).astype(dtype=np.int32)
        df = pd.read_csv("testdata/ahn2_polygon.out", sep=',', header=0, index_col=0, dtype=np.int32)
        assert (pd.DataFrame.equals(df, df_out))

    @staticmethod
    def test_shp_polygons_containsEmpty():
        """ Selecting all points within a Polygon. """
        pc_in = load("testdata/AHN2.las")
        pc_out = select_polygon(pc_in, "testdata/ahn2_geometries_shp/ahn2_polygon_empty.shp", read_from_file=True)
        x = pc_out[point]['x']['data']
        y = pc_out[point]['y']['data']
        assert (len(x) == 0)
        assert (len(y) == 0)


def get_test_data():
    return {'log': ['Processed by module load', 'Processed by module filter using parameters(x,y,z)'],
            'pointcloud': {'offset': {'type': 'float', 'data': 12.1}},
            point:
                {'x': {'type': 'float', 'data': np.array([1.1, 1.2, 1.3])},
                 'y': {'type': 'float', 'data': np.array([2.1, 2.2, 2.3])},
                 'z': {'type': 'float', 'data': np.array([3.1, 3.2, 3.3])},
                 'return': {'type': 'int', 'data': np.array([1, 1, 2])}}}


def select_below_ridiculous_high_z(pc_in):
    return select_below(pc_in, 'z', 100000)


def assert_points_equal(pc_in, pc_out):
    assert_equal(len(pc_out[point]['x']['data']), len(pc_in[point]['x']['data']))
    assert_almost_equal(pc_out[point]['x']['data'], pc_in[point]['x']['data'])
    assert_almost_equal(pc_out[point]['y']['data'], pc_in[point]['y']['data'])
    assert_almost_equal(pc_out[point]['z']['data'], pc_in[point]['z']['data'])
    assert_almost_equal(pc_out[point]['return']['data'], pc_in[point]['return']['data'])


def assert_none_pc_raises_value_error(function, *args, **kwargs):
    with pytest.raises(ValueError):
        pc_in = None
        function(pc_in, *args, **kwargs)


def assert_unknown_key_raises_value_error(function):
    pc = get_test_data()
    with pytest.raises(ValueError):
        function(pc, 'unknown_key_123', 13)


def assert_none_key_raises_value_error(function):
    pc = get_test_data()
    with pytest.raises(ValueError):
        function(pc, None, 13)


def assert_output_is_not_input(function):
    pc_in = get_test_data()
    pc_out = function(pc_in, 'z', 3.2)
    pc_out[point]['x']['data'][0] += 1
    assert_points_equal(pc_in, get_test_data())


def assert_unknown_path_raises_value_error(function):
    pc_in = load("testdata/AHN2.las")
    with pytest.raises(Exception):
        function(pc_in, 'unknown_path_123', read_from_file=True)


def assert_none_path_raises_value_error(function):
    pc_in = load("testdata/AHN2.las")
    with pytest.raises(Exception):
        function(pc_in, None, read_from_file=True)


def assert_none_wkt_raises_value_error(function):
    pc_in = load("testdata/AHN2.las")
    with pytest.raises(ValueError):
        function(pc_in, None)



"""Test read_las module."""
import os
import shutil
import unittest

import numpy as np
import pytest

from laserchicken import keys
from laserchicken.io.las_handler import DEFAULT_LAS_ATTRIBUTES
from laserchicken.io.load import load


class TestReadLas(unittest.TestCase):
    _test_dir = 'TestLoad_dir'
    _test_file_name = 'AHN3.las'
    _test_data_source = 'testdata'
    test_file_path = os.path.join(_test_dir, _test_file_name)

    def test_load_containsPoints(self):
        """Should run without exception and return points."""
        point_cloud = load(self.test_file_path)
        self.assertIn(keys.point, point_cloud)

    def test_load_PointsContainX(self):
        """Should run without exception and return points."""
        point_cloud = load(self.test_file_path)
        print(point_cloud)
        self.assertIn('data', point_cloud[keys.point]['x'])

    def test_load_CorrectFirstX(self):
        """Should run without exception and compare equal."""
        point_cloud = load(self.test_file_path)
        data = {
            'x': 131999.984125,
            'y': 549718.375,
            'z': -0.34100002,
            'gps_time': 78563787.97322202,
            'intensity': 41,
            'raw_classification': 9,
        }
        names = sorted(data)
        print("Order:", names)
        point = [point_cloud[keys.point][name]['data'][0] for name in names]
        np.testing.assert_allclose(np.array(point), np.array([data[name] for name in names]))

    def test_load_nonexistentFile(self):
        """Should raise exception."""
        with pytest.raises(OSError):
            load('nonexistent.las')

    def test_load_defaultAttributes(self):
        point_cloud = load(self.test_file_path)
        expected_attributes = [attr for attr in DEFAULT_LAS_ATTRIBUTES]
        _check_expected_attributes(point_cloud, expected_attributes)

    def test_load_allAttributes(self):
        expected_attributes = ['x', 'y', 'z', 'intensity', 'raw_classification',
                               'scan_angle_rank', 'user_data', 'gps_time',
                               'red', 'green', 'blue']
        for attrs in ('all', ['all']):
            point_cloud = load(self.test_file_path, attributes=attrs)
            for attribute in expected_attributes:
                self.assertIn(attribute, point_cloud[keys.point])

    def test_load_specificAttribute(self):
        """Should return only x,y,z"""
        point_cloud = load(self.test_file_path, attributes=['intensity'])
        expected_attributes = ['x', 'y', 'z', 'intensity']
        _check_expected_attributes(point_cloud, expected_attributes)

    def test_load_noAttributes(self):
        """Should return only x,y,z"""
        point_cloud = load(self.test_file_path, attributes=[])
        expected_attributes = ['x', 'y', 'z']
        _check_expected_attributes(point_cloud, expected_attributes)

    def test_load_invalidAttributes(self):
        """Should raise exception."""
        with pytest.raises(ValueError):
            load(self.test_file_path, attributes=None)
        with pytest.raises(ValueError):
            load(self.test_file_path, attributes=['ytisnetni'])

    def setUp(self):
        os.mkdir(self._test_dir)
        shutil.copyfile(os.path.join(self._test_data_source, self._test_file_name), self.test_file_path)

    def tearDown(self):
        shutil.rmtree(self._test_dir)


class TestReadLaz(TestReadLas):
    _test_dir = 'TestLoad_dir'
    _test_file_name = 'AHN3.laz'
    _test_data_source = 'testdata'
    test_file_path = os.path.join(_test_dir, _test_file_name)


def _check_expected_attributes(point_cloud, attributes):
    for attr in attributes:
        assert attr in point_cloud[keys.point].keys()
    assert [attr for attr in point_cloud[keys.point].keys() if attr not in attributes] == []


"""Test feature extraction."""
import unittest

import numpy as np
from pytest import raises

from laserchicken import keys, test_tools
from laserchicken.feature_extractor import feature_map, feature_extraction
from laserchicken.feature_extractor.mean_std_coeff_feature_extractor import MeanStdCoeffFeatureExtractor
from laserchicken.feature_extractor.median_feature_extractor import MedianFeatureExtractor
from laserchicken.test_feature_extractor import Test1FeatureExtractor
from laserchicken.utils import get_attribute_value
from laserchicken.volume_specification import Sphere
from .feature_test23 import Test2FeatureExtractor, Test3FeatureExtractor, TestVectorizedFeatureExtractor
from .feature_test_broken import TestBrokenFeatureExtractor


class TestExtractFeatures(unittest.TestCase):
    @staticmethod
    def test_extract_single_feature_has_correct_values_in_pc():
        target = test_tools.ComplexTestData().get_point_cloud()
        _compute_features(target, ['test1_a'])
        assert all(target[keys.point]['test1_a']['data'] == 0.5 * target[keys.point]['z']['data'])

    @staticmethod
    def test_extract_only_requested_feature_ends_up_in_pc():
        target = test_tools.ComplexTestData().get_point_cloud()
        _compute_features(target, ['test3_a'])
        assert 'test1_b' not in target[keys.point]

    @staticmethod
    def test_extract_multiple_features_ends_up_in_pc():
        target = test_tools.ComplexTestData().get_point_cloud()
        feature_names = ['test3_a', 'test2_b']
        target = _compute_features(target, feature_names)
        assert ('test3_a' in target[keys.point] and 'test2_b' in target[keys.point])

    @staticmethod
    def test_extract_can_overwrite():
        target = test_tools.ComplexTestData().get_point_cloud()
        target[keys.point]['test1_a'] = {"type": np.float64, "data": [0.9, 0.99, 0.999, 0.9999]}
        feature_names = ['test3_a', 'test1_a']
        target = _compute_features(target, feature_names)
        assert all(target[keys.point]['test1_a']['data'] == 0.5 * target[keys.point]['z']['data'])

    @staticmethod
    def test_extract_unknown_feature():
        with raises(ValueError):
            target = test_tools.ComplexTestData().get_point_cloud()
            _compute_features(target, ['some_unknown_feature'])

    @staticmethod
    def test_vectorized_chunks():
        """Should not throw error for non requested but provided features."""
        n = 2000000  # enough to be too big for a single chunk
        feature_names = ['vectorized1']
        _create_targets_and_extract_features(feature_names, n)

    @staticmethod
    def test_chunks_all_correct_value():
        """Should not throw error for non requested but provided features."""
        n = 2000000  # enough to be too big for a single chunk
        feature_names = ['vectorized1', 'vectorized2', 'test1_a']
        target = _create_targets_and_extract_features(feature_names, n)

        _assert_feature_name_all_valued(1, 'vectorized1', n, target)
        _assert_feature_name_all_valued(1, 'vectorized2', n, target)
        _assert_feature_name_all_valued(0.5, 'test1_a', n, target)

    @staticmethod
    def test_with_neighborhood_generator():
        """Should run for all extractors without error meaning that neighborhood generator is only iterated once.
        Using actual feature extractors here because test feature extractors don't use neighborhoods. """
        n = 200
        feature_names = ['vectorized1', 'test1_a', 'median_z', 'mean_z']
        x = np.ones(n)
        y = np.ones(n)
        z = np.ones(n)
        target = test_tools.create_point_cloud(x, y, z)
        neighborhoods = ([] for _ in range(len(target["vertex"]["x"]["data"])))
        feature_extraction.compute_features({}, neighborhoods, target, feature_names, Sphere(5))

    def setUp(self) -> None:
        self.original_function = feature_map._get_default_extractors
        feature_map._get_default_extractors = _get_test_extractors
        feature_extraction.FEATURES = feature_map.create_default_feature_map()

    def tearDown(self) -> None:
        feature_map._get_default_extractors = self.original_function
        feature_extraction.FEATURES = feature_map.create_default_feature_map()


def _create_targets_and_extract_features(feature_names, n):
    x = np.ones(n)
    y = np.ones(n)
    z = np.ones(n)
    target = test_tools.create_point_cloud(x, y, z)
    _compute_features(target, feature_names)
    return target


def _assert_feature_name_all_valued(expected, feature_name, n, target):
    v = get_attribute_value(target, range(n), feature_name)
    np.testing.assert_allclose(v, expected)


def _compute_features(target, feature_names):
    neighborhoods = [[] for _ in range(len(target["vertex"]["x"]["data"]))]
    feature_extraction.compute_features({}, neighborhoods, target, feature_names, Sphere(5))
    return target


def _get_test_extractors():
    return [Test1FeatureExtractor(), Test2FeatureExtractor(),
            Test3FeatureExtractor(), TestVectorizedFeatureExtractor(),
            TestBrokenFeatureExtractor(), MedianFeatureExtractor(), MeanStdCoeffFeatureExtractor()]



"""Test that the map from feature names to extractor classes is correct."""
import unittest

from laserchicken.feature_extractor import feature_extraction
from laserchicken.test_feature_extractor import Test1FeatureExtractor


class FeatureMapTests(unittest.TestCase):

    def test__feature_map(self):
        expected_features = ['point_density', 'echo_ratio', 'eigenv_1', 'eigenv_2', 'eigenv_3', 'normal_vector_1',
                             'normal_vector_2', 'normal_vector_3', 'slope', 'entropy_z', 'pulse_penetration_ratio',
                             'sigma_z', 'median_z', 'max_z', 'min_z',
                             'range_z', 'var_z', 'mean_z', 'std_z', 'coeff_var_z', 'skew_z', 'kurto_z', 'skew_normalized_height',
                             'mean_normalized_height','std_normalized_height', 'coeff_var_normalized_height', 'var_normalized_height', 'min_normalized_height',
                             'max_normalized_height', 'range_normalized_height', 'kurto_normalized_height', 'entropy_normalized_height',
                             'median_normalized_height', 'max_intensity',
                             'density_absolute_mean_z', 'density_absolute_mean_normalized_height', 'perc_15_z',
                             'perc_99_normalized_height', 'max_intensity', 'min_intensity', 'range_intensity', 'mean_intensity', 'std_intensity', 'coeff_var_intensity']
        for feature in expected_features:
            self.assertIn(feature, feature_extraction.FEATURES)

    def test_feature_map_contains_new_feature_after_registration(self):
        test_feature_extractor = Test1FeatureExtractor()
        expected_features = test_feature_extractor.provides()

        feature_extraction.register_new_feature_extractor(test_feature_extractor)

        for feature in expected_features:
            self.assertIn(feature, feature_extraction.FEATURES)

    def test_list_all_feature_names_contains_items(self):
        names = feature_extraction.list_feature_names()
        self.assertGreater(len(names), 0)

    def test_list_all_feature_names_are_strings(self):
        names = feature_extraction.list_feature_names()
        for name in names:
            self.assertEqual(type(name), str)



import os
import shutil
import unittest

import numpy as np
import pylas
import pytest

from laserchicken import keys
from laserchicken.test_tools import SimpleTestData, ComplexTestData
from laserchicken.test_tools import create_point_cloud
from laserchicken.io.export import export
from laserchicken.io.load import load


class TestWriteLas(unittest.TestCase):
    _test_dir = 'TestLoad_dir'
    _test_file_name = 'test.las'
    _load_file_name = 'AHN3.las'
    _load_data_source = 'testdata'
    test_file_path = os.path.join(_test_dir, _test_file_name)
    load_file_path = os.path.join(_load_data_source, _load_file_name)

    def test_write_emptyData(self):
        """ Should raise an exception """
        pc = create_point_cloud([], [], [])
        with pytest.raises(ValueError):
            export(pc, self.test_file_path)

    def test_write_nonExistentDir(self):
        """ Should raise an exception """
        pc = SimpleTestData.get_point_cloud()
        with pytest.raises(FileNotFoundError):
            export(pc, os.path.join('/nonexistentdir', self._test_file_name))

    def test_write_sameFileTwice(self):
        """ Should raise an exception """
        pc = SimpleTestData.get_point_cloud()
        export(pc, self.test_file_path)
        with pytest.raises(FileExistsError):
            export(pc, self.test_file_path)

    def test_write_sameFileTwiceOverwrite(self):
        """ Should not raise an exception """
        pc = SimpleTestData.get_point_cloud()
        export(pc, self.test_file_path)
        export(pc, self.test_file_path, overwrite=True)
        self.assertTrue(os.path.isfile(self.test_file_path))

    def test_write_loadTheSameSimpleData(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        pc = SimpleTestData.get_point_cloud()
        export(pc, self.test_file_path)
        file = _get_file_from_path(self.test_file_path)
        _assert_all_attributes_in_file(pc[keys.point], file)

    def test_write_loadTheSameComplexData(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        test_data = ComplexTestData()
        pc = test_data.get_point_cloud()
        export(pc, self.test_file_path)
        file = _get_file_from_path(self.test_file_path)
        _assert_all_attributes_in_file(pc[keys.point], file)

    def test_check_is_compressed(self):
        """ Writing LAS file should not generate compressed file. """
        test_data = ComplexTestData()
        pc = test_data.get_point_cloud()
        export(pc, self.test_file_path)
        file = _get_file_from_path(self.test_file_path)
        self.assertFalse(file.header.are_points_compressed)

    def test_write_invalidAttributes(self):
        """ Should raise exception. """
        test_data = ComplexTestData()
        pc = test_data.get_point_cloud()
        with pytest.raises(ValueError):
            export(pc, self.test_file_path, attributes=None)
        with pytest.raises(ValueError):
            export(pc, self.test_file_path, attributes=['ytisnetni'])

    def test_write_loadRealData(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        pc = load(self.load_file_path)
        export(pc, self.test_file_path)
        file = _get_file_from_path(self.test_file_path)
        _assert_all_attributes_in_file(pc[keys.point], file)

    def test_write_processedRealData(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        pc = load(self.load_file_path)
        x = pc[keys.point]['x']['data']
        pc[keys.point]['test_feature'] = {'data': np.zeros_like(x), 'type': 'float64'}
        export(pc, self.test_file_path)
        file = _get_file_from_path(self.test_file_path)
        _assert_all_attributes_in_file(pc[keys.point], file)

    def setUp(self):
        os.mkdir(self._test_dir)

    def tearDown(self):
        shutil.rmtree(self._test_dir)


class TestWriteLaz(TestWriteLas):
    _test_dir = 'TestLoad_dir'
    _test_file_name = 'test.laz'
    _load_file_name = 'AHN3.laz'
    _load_data_source = 'testdata'
    test_file_path = os.path.join(_test_dir, _test_file_name)
    load_file_path = os.path.join(_load_data_source, _load_file_name)

    def test_check_is_compressed(self):
        """ Writing LAZ file should generate compressed file. """
        test_data = ComplexTestData()
        pc = test_data.get_point_cloud()
        export(pc, self.test_file_path)
        file = _get_file_from_path(self.test_file_path)
        self.assertTrue(file.header.are_points_compressed)


def _assert_all_attributes_in_file(attributes, file):
    for name, attribute in attributes.items():
        data = attribute['data']
        dtype = np.dtype(attribute['type'])
        assert hasattr(file, name)
        file_data = getattr(file, name)
        assert dtype.name == file_data.dtype.name
        assert data.size == file_data.size
        np.testing.assert_allclose(data, file_data, err_msg="{} differ".format(name))


def _get_file_from_path(path):
    return pylas.read(path)



import os
import shutil
import unittest

import pytest

from laserchicken.test_tools import SimpleTestData, ComplexTestData
from laserchicken.io.export import export


def read_header(ply, is_binary=False):
    header = ''
    line = ply.readline()
    if is_binary: line = line.decode("utf-8")
    while line.strip() != 'end_header':
        header = header + line
        line = ply.readline()
        if is_binary: line = line.decode("utf-8")
    return header


def read_data_ascii(ply):
    data = ''
    in_header = True
    for line in ply:
        if line == 'end_header\n':
            in_header = False
        else:
            if not in_header:
                data = data + line
    return data


def read_data_binary(ply):
    data = b''
    in_header = True
    for line in ply:
        if in_header:
            if line.decode('utf-8') == 'end_header\n':
                in_header = False
        else:
            if not in_header:
                data = data + line.rstrip()
    return data


class TestWritePly(unittest.TestCase):
    _test_dir = 'TestLoad_dir'
    _test_file_name = 'test.ply'
    _test_data_source = 'testdata'
    test_file_path = os.path.join(_test_dir, _test_file_name)

    def test_write_nonExistingFile(self):
        """ Should create a file. """
        pc = SimpleTestData.get_point_cloud()
        export(pc, self.test_file_path)
        assert (os.path.exists(self.test_file_path))

    def test_write_sameFileTwice(self):
        """ Should throw an exception. """
        pc = SimpleTestData.get_point_cloud()
        export(pc, self.test_file_path)
        # Catch most specific subclass of FileExistsError (3.6) and IOError (2.7).
        with pytest.raises(Exception):
            export(pc, self.test_file_path)

    def test_write_sameFileTwiceOverwrite(self):
        """ Should not raise an exception """
        pc = SimpleTestData.get_point_cloud()
        export(pc, self.test_file_path)
        export(pc, self.test_file_path, overwrite=True)
        self.assertTrue(os.path.isfile(self.test_file_path))

    def test_write_invalidAttributes(self):
        """ Should raise exception. """
        test_data = ComplexTestData()
        pc = test_data.get_point_cloud()
        with pytest.raises(ValueError):
            export(pc, self.test_file_path, attributes=None)
        with pytest.raises(ValueError):
            export(pc, self.test_file_path, attributes=['ytisnetni'])

    def test_write_loadTheSameSimpleHeader(self):
        """  Writing a simple point cloud and loading it afterwards should result in the same point cloud."""
        pc_in = SimpleTestData.get_point_cloud()
        header_in = SimpleTestData.get_header()
        export(pc_in, self.test_file_path)
        with open(self.test_file_path, 'r') as ply:
            header_out = read_header(ply)
        self.assertMultiLineEqual(header_in, header_out)

    def test_write_loadTheSameComplexHeader(self):
        """  Writing a complex point cloud and loading it afterwards should result in the same point cloud."""
        test_data = ComplexTestData()
        pc_in = test_data.get_point_cloud()
        header_in = test_data.get_header()
        export(pc_in, self.test_file_path)
        with open(self.test_file_path, 'r') as ply:
            header_out = read_header(ply)
        self.assertMultiLineEqual(header_in, header_out)

    def test_write_loadTheSameSimpleData(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        pc_in = SimpleTestData.get_point_cloud()
        export(pc_in, self.test_file_path)
        data_in = SimpleTestData.get_data()
        with open(self.test_file_path, 'r') as ply:
            data_out = read_data_ascii(ply)
        self.assertEqual(data_in, data_out)

    def test_write_loadTheSameComplexData(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        test_data = ComplexTestData()
        pc_in = test_data.get_point_cloud()
        export(pc_in, self.test_file_path)
        data_in = test_data.get_data()
        with open(self.test_file_path, 'r') as ply:
            data_out = read_data_ascii(ply)
        self.assertEqual(data_in, data_out)

    def test_write_loadTheSameSimpleHeaderBinary(self):
        """  Writing a simple point cloud and loading it afterwards should result in the same point cloud."""
        pc_in = SimpleTestData.get_point_cloud()
        header_in = SimpleTestData.get_header(is_binary=True)
        export(pc_in, self.test_file_path, is_binary=True)
        with open(self.test_file_path, 'rb') as ply:
            header_out = read_header(ply, is_binary=True)
        self.assertMultiLineEqual(header_in, header_out)

    def test_write_loadTheSameSimpleDataBinary(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        pc_in = SimpleTestData.get_point_cloud()
        export(pc_in, self.test_file_path, is_binary=True)
        data_in = SimpleTestData.get_data(is_binary=True)
        with open(self.test_file_path, 'rb') as ply:
            data_out = read_data_binary(ply)
        self.assertEqual(data_in, data_out)

    def test_write_loadTheSameComplexHeaderBinary(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        test_data = ComplexTestData()
        pc_in = test_data.get_point_cloud()
        header_in = test_data.get_header(is_binary=True)
        export(pc_in, self.test_file_path, is_binary=True)
        with open(self.test_file_path, 'rb') as ply:
            header_out = read_header(ply, is_binary=True)
        self.assertEqual(header_in, header_out)

    def test_write_loadTheSameComplexDataBinary(self):
        """ Writing point cloud data and loading it afterwards should result in the same point cloud data. """
        test_data = ComplexTestData()
        pc_in = test_data.get_point_cloud()
        data_in = test_data.get_data(is_binary=True)
        export(pc_in, self.test_file_path, is_binary=True)
        with open(self.test_file_path, 'rb') as ply:
            data_out = read_data_binary(ply)
        self.assertEqual(data_in, data_out)

    def setUp(self):
        os.mkdir(self._test_dir)

    def tearDown(self):
        shutil.rmtree(self._test_dir)



import os

from setuptools import setup, find_packages


def read(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


required = read('requirements.txt').splitlines()
version = read('laserchicken/_version.txt').strip()


setup(
    name='laserchicken',
    version=version,
    description='Point cloud toolkit',
    license='Apache 2.0',
    keywords=['Python', 'Point cloud'],
    url='https://github.com/eEcoLiDAR/eEcoLiDAR',
    packages=find_packages(),
    install_requires=required,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ],
    entry_points={
        'console_scripts': [
            'laserchicken = laserchicken.tools.cli:main',
        ],
    },
    package_data={'': ['_version.txt']},
)



import sys
import copy
import datetime

import numpy as np

from laserchicken import keys, _version


def get_point(point_cloud, index):
    """
    Get x, y, z tuple of one or more points in a point cloud.

    :param point_cloud: point cloud containing the point of interest
    :param index: index of the point within the point cloud
    :return: x, y, z as a tuple of floats
    """
    return point_cloud[keys.point]["x"]["data"][index], point_cloud[keys.point]["y"]["data"][index], \
           point_cloud[keys.point]["z"]["data"][index]


def get_xyz_per_neighborhood(sourcepc, neighborhoods):
    """
    Get x, y, z tuple for each point in a neighborhood for each neighborhood.
    :param sourcepc:
    :param neighborhoods:
    :return: 3d tensor as a masked array
    """
    max_length = max(map(lambda x: len(x), neighborhoods))

    xyz_grp = np.zeros((len(neighborhoods), 3, max_length))
    mask = np.zeros((len(neighborhoods), 3, max_length))
    for i, neighborhood in enumerate(neighborhoods):
        n_neighbors = len(neighborhood)
        if n_neighbors == 0:
            continue
        x, y, z = get_point(sourcepc, neighborhood)
        xyz_grp[i, 0, :n_neighbors] = x
        xyz_grp[i, 1, :n_neighbors] = y
        xyz_grp[i, 2, :n_neighbors] = z
        mask[i, :, :n_neighbors] = 1
    return np.ma.MaskedArray(xyz_grp, mask == 0)


def get_attributes_per_neighborhood(point_cloud, neighborhoods, attribute_names):
    """
    Get attribute values for each point in a neighborhood for each neighborhood.
    :param point_cloud:
    :param neighborhoods:
    :param attribute_names: list of attribute names
    :return: 3d tensor as a masked array
    """
    max_length = max(map(lambda x: len(x), neighborhoods))

    xyz_grp = np.zeros((len(neighborhoods), (len(attribute_names)), max_length))
    mask = np.zeros((len(neighborhoods), (len(attribute_names)), max_length))
    for i_neighborhood, neighborhood in enumerate(neighborhoods):
        n_neighbors = len(neighborhood)
        if n_neighbors == 0:
            continue
        for i_attribute, attribute_name in enumerate(attribute_names):
            attribute = get_attribute_value(point_cloud, neighborhood, attribute_name)
            xyz_grp[i_neighborhood, i_attribute, :n_neighbors] = attribute
        mask[i_neighborhood, :, :n_neighbors] = 1
    return np.ma.MaskedArray(xyz_grp, mask == 0)


def get_attribute_value(point_cloud, index, attribute_name):
    """
    Get value of a single attribute of a single point in a point cloud.

    :param point_cloud: point cloud containing the point of interest
    :param index: index of the point within the point cloud
    :param attribute_name: attribute name
    :return: value of the attribute of the point
    """
    return point_cloud[keys.point][attribute_name]["data"][index]


def get_features(point_cloud, attribute_names, index=None):
    """
    Get value of each attribute in a list for a single point in a point cloud.

    :param point_cloud: point cloud containing the point of interest
    :param attribute_names: attribute names
    :param index: index of the point within the point cloud
    :return: list of values of the attributes of the point
    """
    if index is None:
        index = list(range(point_cloud[keys.point]['x']["data"].shape[0]))
    return (point_cloud[keys.point][f]["data"][index] for f in attribute_names)


def create_point_cloud(x, y, z):
    """
    Create a point cloud object given only the x y z values.

    :param x: x attribute values
    :param y: y attribute values
    :param z: z attribute values
    :return: point cloud object
    """
    return {keys.point: {'x': {'type': 'float64', 'data': np.array(x)},
                         'y': {'type': 'float64', 'data': np.array(y)},
                         'z': {'type': 'float64', 'data': np.array(z)}},
            keys.point_cloud: {},
            keys.provenance: []
            }


def copy_point_cloud(source_point_cloud, array_mask=None):
    """
    Makes a deep copy of a point cloud dict using the array mask when copying the points.

    :param source_point_cloud: Input point cloud
    :param array_mask: A mask indicating which points to copy.
    :return: The copy including only the masked points.
    """
    result = {}
    for key, value in source_point_cloud.items():
        if isinstance(value, dict):
            new_value = copy_point_cloud(value, array_mask)
        elif isinstance(value, np.ndarray):
            if array_mask is not None:
                new_value = value[array_mask] if any(value) else np.copy(value)
            else:
                new_value = np.copy(value)
        else:
            new_value = copy.copy(value)
        result[key] = new_value
    return result


def add_metadata(point_cloud, module, params):
    """
    Adds module metadata to point cloud provenance

    :param point_cloud:
    :param module:
    :param params:
    :return:
    """
    msg = {"time": str(datetime.datetime.utcnow()),
           "module": module.__name__ if hasattr(module, "__name__") else str(module)}
    if any(params):
        msg["parameters"] = params
    msg["version"] = _version.__version__
    if keys.provenance not in point_cloud:
        point_cloud[keys.provenance] = []
    point_cloud[keys.provenance].append(msg)


def add_to_point_cloud(point_cloud_1, point_cloud_2, add_log=True):
    """
    Add points from a point cloud to another

    :param point_cloud_1: point cloud where points are added (it can be empty)
    :param point_cloud_2: point cloud from which points are taken
    :param add_log: whether to add a log to the point cloud structure
    :return: updated point cloud
    """
    if keys.point in point_cloud_1:
        # check whether second point cloud is valid
        if keys.point not in point_cloud_2:
            raise TypeError('Invalid point cloud provided!')

        # if first point cloud is empty, fill it with attributes of second point cloud
        if len(point_cloud_1[keys.point]['x']['data']) == 0:
            for key, value in copy_point_cloud(point_cloud_2).items():
                point_cloud_1[key] = value
            return point_cloud_1
    else:
        # down the tree structure, the point clouds need to have the same attributes
        attributes_1 = sorted(point_cloud_1.keys())
        attributes_2 = sorted(point_cloud_2.keys())
        if attributes_1 != attributes_2:
            raise AttributeError('Attributes differ: [{}] <-> [{}]'.format(", ".join(attributes_1),
                                                                           ", ".join(attributes_2)))

    for key, value in point_cloud_2.items():
        # if root attributes are missing (e.g. log), add them
        if key not in point_cloud_1:
            point_cloud_1[key] = copy.copy(value)
            continue

        # check type of data to merge
        if not isinstance(point_cloud_1[key], type(value)):
            raise TypeError('Types of attribute {} differ in the two point clouds: '
                            '{} <-> {}'.format(key, type(point_cloud_1[key]), type(value)))

        if isinstance(value, dict):
            add_to_point_cloud(point_cloud_1[key], value, add_log=False)
        elif isinstance(value, list):
            point_cloud_1[key] += value
        elif isinstance(value, np.ndarray):
            point_cloud_1[key] = np.concatenate((point_cloud_1[key], value))
        else:
            # non extendable elements should be identical
            if point_cloud_1[key] != value:
                raise ValueError('Point clouds differ in attribute {}: '
                                 '{} <-> {}'.format(key, point_cloud_1[key], value))
    if add_log:
        add_metadata(point_cloud_1, sys.modules[__name__], 'point clouds merged')
    return point_cloud_1


def fit_plane_svd(xpts, ypts, zpts):
    """
    Fit a plane to a series of points given as x,y,z coordinates.
    
    r=Return the normal vector to the plane
    Use the SVD methods described for example here
    https://www.ltu.se/cms_fs/1.51590!/svd-fitting.pdf

    :param x: x coordinate of the points
    :param y: y coordinate of the points
    :param z: z coordinate of the points
    :return: normal vector of the plane
    """
    # check size consistency
    if xpts.size != ypts.size or xpts.size != zpts.size or ypts.size != zpts.size:
        raise AssertionError("coordinate size don't match")
    npts = xpts.size

    # form the A matrix of the coordinate
    a = np.column_stack((xpts, ypts, zpts))
    a -= np.sum(a, 0) / npts

    # compute the SVD
    u, _, _ = np.linalg.svd(a.T)

    # return the normal vector
    return u[:, 2]


def fit_plane(x, y, a):
    """
    Fit a plane and return a function that returns a for every given x and y.

    Solves Ax = b where A is the matrix of (x,y) combinations, x are the plane parameters, and b the values.
    Example:
    >>> points = np.random.rand(100, 3)
    >>> f = fit_plane(points[:, 0], points[:, 1], points[:, 2])
    >>> new_points = np.random.rand(10, 3)
    >>> f(new_points[0], new_points [1])

    :param x: x coordinates
    :param y: y coordinates
    :param a: value (for instance height)
    :return: function that returns a for every given x and y
    """
    matrix = np.column_stack((np.ones(x.size), x, y))
    parameters, _, _, _ = np.linalg.lstsq(matrix, a)
    return lambda x_in, y_in: np.stack((np.ones(len(x)), x_in, y_in)).T.dot(parameters)


def update_feature(point_cloud, feature_name, value, array_mask=None, add_log=True):
    """
    Update one feature of the point cloud and assign value.
    The feature 
    If the feature does not exist in the point cloud, add it.

    :param point_cloud: point cloud to update the feature
    :param feature_name: name of the feature
    :param value: value of the feature. Can be a signl value or an array
    :param array_mask: A mask indicating which point to update the feature
    :param add_log: whether to add a log to the point cloud structure
    :return: updated point cloud
    """

    # Get and check input data type
    if isinstance(value, np.ndarray):
        data_type = value.dtype.name
    else:
        if isinstance(value, (str, int, float, bool)):
            data_type = type(value).__name__
        else:
           raise TypeError("value must be numpy ndarray, or one in (str, int, float, bool)") 
    
    # Check mask size and type
    if array_mask is not None:
        if array_mask.size != len(point_cloud[keys.point]['x']['data']):
            raise AssertionError("Mask size: {} doesn't match the size of point cloud column: {}".format(array_mask.size, len(point_cloud[keys.point]['x']['data'])))

    # If value is ndarray, it shall have the same length as x, or true elements in mask
    if isinstance(value, np.ndarray):
        if array_mask is None:
            if value.size != len(point_cloud[keys.point]['x']['data']): 
                raise AssertionError("value size: {} doesn't match the size of point cloud column: {}".format(value.size, len(point_cloud[keys.point]['x']['data'])))
        else:
            if value.size != np.sum(array_mask): 
                raise AssertionError("value size: {} doesn't match the number of elements in mask: {}".format(value.size, np.sum(array_mask)))
    
    # Check if the feature exists if not create the column
    if not feature_name in point_cloud[keys.point]:
        point_cloud[keys.point][feature_name] = {'data':np.zeros(len(point_cloud[keys.point]['x']['data']), dtype=data_type),
                                                 'type':data_type}
    
    # Convert data of the feature if necccesary
    if point_cloud[keys.point][feature_name]['type'] != data_type:
        print('Setting data type of {} as {}'.format(feature_name, data_type))
        point_cloud[keys.point][feature_name]['data'] = point_cloud[keys.point][feature_name]['data'].astype(data_type)
        point_cloud[keys.point][feature_name]['type'] = data_type

    # Update the column 
    if array_mask is None:
        point_cloud[keys.point][feature_name]['data'][:] = value
    else:
        point_cloud[keys.point][feature_name]['data'][array_mask] = value

    if add_log:
        add_metadata(point_cloud, sys.modules[__name__], 'add feature {} to point cloud.'.format(feature_name))

    return point_cloud


from laserchicken import keys
from laserchicken.feature_extractor.band_ratio_feature_extractor import BandRatioFeatureExtractor
from .density_absolute_mean_feature_extractor import DensityAbsoluteMeanFeatureExtractor
from .density_feature_extractor import PointDensityFeatureExtractor
from .echo_ratio_feature_extractor import EchoRatioFeatureExtractor
from .eigenvals_feature_extractor import EigenValueVectorizeFeatureExtractor
from .entropy_feature_extractor import EntropyFeatureExtractor
from .kurtosis_feature_extractor import KurtosisFeatureExtractor
from .mean_std_coeff_feature_extractor import MeanStdCoeffFeatureExtractor
from .median_feature_extractor import MedianFeatureExtractor
from .percentile_feature_extractor import PercentileFeatureExtractor
from .pulse_penetration_feature_extractor import PulsePenetrationFeatureExtractor
from .range_feature_extractor import RangeFeatureExtractor
from .sigma_z_feature_extractor import SigmaZFeatureExtractor
from .skew_feature_extractor import SkewFeatureExtractor
from .var_feature_extractor import VarianceFeatureExtractor


def create_default_feature_map():
    """Construct a mapping from feature names to feature extractor classes."""
    extractors = _get_default_extractors()
    name_extractor_pairs = _create_name_extractor_pairs(extractors)
    return {feature_name: extractor for feature_name, extractor in name_extractor_pairs}


def _create_name_extractor_pairs(extractors=None):
    if extractors is None:
        extractors = _get_default_extractors()
    name_extractor_pairs = [(feature_name, extractor)
                            for extractor in extractors
                            for feature_name in extractor.provides()]
    return name_extractor_pairs


def _get_default_extractors():
    return [PointDensityFeatureExtractor(),
            EchoRatioFeatureExtractor(),
            EigenValueVectorizeFeatureExtractor(),
            EntropyFeatureExtractor(),
            EntropyFeatureExtractor(data_key=keys.normalized_height),
            PulsePenetrationFeatureExtractor(),
            SigmaZFeatureExtractor(),
            MedianFeatureExtractor(),
            MedianFeatureExtractor(data_key=keys.normalized_height),
            VarianceFeatureExtractor(),
            VarianceFeatureExtractor(data_key=keys.normalized_height),
            MeanStdCoeffFeatureExtractor(),
            MeanStdCoeffFeatureExtractor(data_key=keys.normalized_height),
            MeanStdCoeffFeatureExtractor(data_key=keys.intensity),
            SkewFeatureExtractor(),
            SkewFeatureExtractor(data_key=keys.normalized_height),
            KurtosisFeatureExtractor(),
            KurtosisFeatureExtractor(data_key=keys.normalized_height),
            RangeFeatureExtractor(),
            RangeFeatureExtractor(data_key=keys.normalized_height),
            RangeFeatureExtractor(data_key=keys.intensity),
            DensityAbsoluteMeanFeatureExtractor(),
            DensityAbsoluteMeanFeatureExtractor(data_key=keys.normalized_height),
            BandRatioFeatureExtractor(None, 1, data_key=keys.normalized_height),
            BandRatioFeatureExtractor(1, 2, data_key=keys.normalized_height),
            BandRatioFeatureExtractor(2, 3, data_key=keys.normalized_height),
            BandRatioFeatureExtractor(3, None, data_key=keys.normalized_height)] \
           + [PercentileFeatureExtractor(percentile=p) for p in range(1, 101)] \
           + [PercentileFeatureExtractor(percentile=p, data_key=keys.normalized_height) for p in range(1, 101)]



"""Calculate echo ratio.

See https://github.com/eEcoLiDAR/eEcoLiDAR/issues/21
"""

import numpy as np

from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.keys import point
from laserchicken.utils import get_xyz_per_neighborhood, get_point


class EchoRatioFeatureExtractor(FeatureExtractor):
    """Feature extractor for the point density."""

    @classmethod
    def requires(cls):
        """
        Get a list of names of the point attributes that are needed for this feature extraction.

        For simple features, this could be just x, y, and z. Other features can build on again
        other features to have been computed first.

        :return: List of feature names
        """
        return []

    @classmethod
    def provides(cls):
        """
        Get a list of names of the feature values.

        This will return as many names as the number feature values that will be returned.
        For instance, if a feature extractor returns the first 3 Eigen values, this method
        should return 3 names, for instance 'eigen_value_1', 'eigen_value_2' and 'eigen_value_3'.

        :return: List of feature names
        """
        return ['echo_ratio']

    def extract(self, point_cloud, neighborhoods, target_point_cloud, target_index, volume_description):
        """
        Extract the feature value(s) of the point cloud at location of the target.

        :param point_cloud: environment (search space) point cloud
        :param neighborhoods: array of array of indices of points within the point_cloud argument
        :param target_point_cloud: point cloud that contains target point
        :param target_index: index of the target point in the target point cloud
        :param volume_description: volume object that describes the shape and size of the search volume
        :return: feature value
        """
        if volume_description.TYPE != 'infinite cylinder':
            raise ValueError('The volume must be a cylinder')

        if target_point_cloud is None:
            raise ValueError('Target point cloud required')

        if target_index is None:
            raise ValueError('Target point index required')

        # xyz = self.get_neighborhood_positions(point_cloud, neighborhood)
        xyz = get_xyz_per_neighborhood(point_cloud, neighborhoods)
        n_cylinder = np.sum(xyz.mask[:, 0, :] == False, axis=1, dtype=float)

        x0, y0, z0 = get_point(target_point_cloud, target_index)
        xyz0 = np.column_stack((x0, y0, z0))

        difference = xyz - xyz0[:, :, None]
        sum_of_squares = np.sum(difference ** 2, 1)
        n_sphere = np.sum(sum_of_squares <= volume_description.radius ** 2, axis=1)
        return n_sphere / n_cylinder

    @staticmethod
    def get_target_positions(target_point_cloud, target_index):
        x0 = target_point_cloud[point]['x']['data'][target_index]
        y0 = target_point_cloud[point]['y']['data'][target_index]
        z0 = target_point_cloud[point]['z']['data'][target_index]
        return np.array([x0, y0, z0])

    @staticmethod
    def get_neighborhood_positions(point_cloud, neighborhood):
        x = point_cloud[point]['x']['data'][neighborhood]
        y = point_cloud[point]['y']['data'][neighborhood]
        z = point_cloud[point]['z']['data'][neighborhood]
        return np.column_stack((x, y, z))

    def get_params(self):
        """
        Return a tuple of parameters involved in the current feature extractorobject.

        Needed for provenance.
        """
        return ()



import os
import random
import time
import unittest

import numpy as np

from laserchicken import compute_neighbors, keys, load, utils
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor
from laserchicken.feature_extractor.feature_extraction import compute_features
from laserchicken.test_tools import create_point_cloud
from laserchicken.utils import copy_point_cloud
from laserchicken.volume_specification import InfiniteCylinder
from .eigenvals_feature_extractor import EigenValueVectorizeFeatureExtractor


class TestExtractEigenValues(unittest.TestCase):
    def test_eigenvalues_in_cylinders(self):
        """Test provenance added (This should actually be part the general feature extractor test suite)."""
        random.seed(102938482634)
        point_cloud = load(os.path.join('testdata', 'AHN3.las'))
        num_all_pc_points = len(point_cloud[keys.point]["x"]["data"])
        rand_indices = [random.randint(0, num_all_pc_points) for _ in range(20)]
        target_point_cloud = utils.copy_point_cloud(point_cloud, rand_indices)
        radius = 2.5
        neighbors = compute_neighbors.compute_cylinder_neighborhood(point_cloud, target_point_cloud, radius)

        compute_features(point_cloud, neighbors, target_point_cloud,
                         ["eigenv_1", "eigenv_2", "eigenv_3"], InfiniteCylinder(5))

        self.assertEqual("laserchicken.feature_extractor.eigenvals_feature_extractor",
                         target_point_cloud[keys.provenance][-1]["module"])

    @staticmethod
    def test_eigenvalues_of_too_few_points_results_in_0():
        """If there are too few points to calculate the eigen values don't output NaN or inf."""
        a = np.array([5])
        pc = create_point_cloud(a, a, a)

        compute_features(pc, [[0]], pc, ["eigenv_1", "eigenv_2", "eigenv_3"], InfiniteCylinder(5))

        eigen_val_123 = np.array([pc[keys.point]['eigenv_{}'.format(i)]['data'] for i in [1, 2, 3]])
        assert not np.any(np.isnan(eigen_val_123))
        assert not np.any(np.isinf(eigen_val_123))


class TestExtractEigenvaluesComparison(unittest.TestCase):
    point_cloud = None

    def test_eigen_multiple_neighborhoods(self):
        """
        Test and compare the serial and vectorized eigenvalues.

        Eigenvalues are computed for a list of neighborhoods in real data. A vectorized implementation and a serial
        implementation are compared and timed. Any difference in result between the two methods is definitely
        unexpected (except maybe in ordering of eigen values).
        """
        # vectorized version
        t0 = time.time()
        extract_vect = EigenValueVectorizeFeatureExtractor()
        eigvals_vect = extract_vect.extract(self.point_cloud, self.neigh, None, None, None)
        print('Timing Vectorize : {}'.format((time.time() - t0)))
        eigvals_vect = np.vstack(eigvals_vect[:3]).T

        # serial version
        eigvals = []
        t0 = time.time()
        for n in self.neigh:
            extract = EigenValueSerial()
            eigvals.append(extract.extract(self.point_cloud, n, None, None, None))
        print('Timing Serial : {}'.format((time.time() - t0)))
        eigvals = np.array(eigvals)

        np.testing.assert_allclose(eigvals_vect, eigvals)

    def setUp(self):
        """
        Set up the test.

        Load in a bunch of real data from AHN3.
        """
        np.random.seed(1234)

        _TEST_FILE_NAME = 'AHN3.las'
        _TEST_DATA_SOURCE = 'testdata'

        _CYLINDER = InfiniteCylinder(4)
        _PC_260807 = load(os.path.join(_TEST_DATA_SOURCE, _TEST_FILE_NAME))
        _PC_1000 = copy_point_cloud(_PC_260807, array_mask=(
            np.random.choice(range(len(_PC_260807[keys.point]['x']['data'])), size=1000, replace=False)))
        _1000_NEIGHBORHOODS_IN_260807 = list(compute_neighbors.compute_neighborhoods(_PC_260807, _PC_1000, _CYLINDER))

        self.point_cloud = _PC_260807
        self.neigh = _1000_NEIGHBORHOODS_IN_260807


class TestExtractNormalPlaneArtificialData0(unittest.TestCase):
    def test_regression(self):
        n_x = 3
        slope_ = 1.0
        x_, y_ = np.meshgrid(range(n_x), range(n_x))
        x = x_.flatten()
        y = y_.flatten()
        z = [slope_ * x[i] for i in range(len(x))]

        dim = (1, 3, len(x))
        mask = np.zeros(dim)
        xyz_grp = np.ma.MaskedArray(np.zeros(dim), mask == 0)
        xyz_grp[0, 0, :] = x
        xyz_grp[0, 1, :] = y
        xyz_grp[0, 2, :] = z

        point_cloud = create_point_cloud(xyz_grp[0, 0, :], xyz_grp[0, 1, :], xyz_grp[0, 2, :])
        extractor = EigenValueVectorizeFeatureExtractor()
        neighborhoods = [list(range(len(x)))]
        normals = np.array(extractor.extract(point_cloud, neighborhoods, point_cloud, None, None)[3:6])[:, 0]
        np.testing.assert_allclose(normals, np.array((-np.sqrt(2) / 2, 0, np.sqrt(2) / 2)))

    def test_from_eigen(self):
        nvect = np.array([1., 2., 3.])
        neighborhood, pc = create_point_cloud_in_plane_and_neighborhood(nvect)
        extractor = EigenValueVectorizeFeatureExtractor()

        n1, n2, n3, slope_fit = extractor.extract(pc, neighborhood, None, None, None)[3:]

        np.testing.assert_allclose(nvect[0], n1[0])
        np.testing.assert_allclose(nvect[1], n2[0])
        np.testing.assert_allclose(nvect[2], n3[0])

    def test_normal_always_up(self):
        """Tests whether resulting normals are always pointing upwards (positive z component). As this should happen
        by chance a lot already, we test many times to make sure results are positive consistently."""
        z_of_normals = []
        for i in range(100):
            neighborhood, pc = create_point_cloud_in_plane_and_neighborhood()
            z_of_normals += list(EigenValueVectorizeFeatureExtractor().extract(pc, neighborhood, None, None, None)[5])
        np.testing.assert_array_less(np.zeros_like(z_of_normals), z_of_normals)

    def test_normal_unit_length(self):
        """Tests whether resulting normals are unit length"""
        neighborhood, pc = create_point_cloud_in_plane_and_neighborhood()
        normals = np.array(EigenValueVectorizeFeatureExtractor().extract(pc, neighborhood, None, None, None)[3:6])
        lengths = np.sum(normals * normals, axis=0)
        np.testing.assert_almost_equal(np.ones_like(lengths), lengths)


class TestExtractSlopeArtificialData(unittest.TestCase):
    def test_positive_slope(self):
        """Tests whether resulting slopes are always positive. As this should happen
        by chance a lot already, we test many times to make sure results are positive consistently."""
        slopes = []
        for i in range(100):
            neighborhood, pc = create_point_cloud_in_plane_and_neighborhood()
            slopes += list(EigenValueVectorizeFeatureExtractor().extract(pc, neighborhood, None, None, None)[6])
        np.testing.assert_array_less(np.zeros_like(slopes), slopes)

    def test_001_has_slope_0(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 0., 1.]), 0.)

    def test_011_has_slope_1(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 1., 1.]), 1.)

    def test_01min1_has_slope_1(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 1., -1.]), 1.)

    def test_0min11_has_slope_1(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., -1., 1.]), 1.)

    def test_00251_has_slope_025(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 0.25, 1.]), 0.25)

    def test_021_has_slope_2(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 2., 1.]), 2.)

    def test_031_has_slope_3(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 3., 1.]), 3.)

    def test_0205_has_slope_4(self):
        self.assert_data_with_normal_vector_has_slope(np.array([0., 2., 0.5]), 4.)

    @staticmethod
    def assert_data_with_normal_vector_has_slope(nvect, expected_slope):
        """Create random point cloud with normal vector nvect and assert that estimated slope has expected value."""
        neighborhood, pc = create_point_cloud_in_plane_and_neighborhood(nvect)
        extractor = EigenValueVectorizeFeatureExtractor()
        slope = extractor.extract(pc, neighborhood, None, None, None)[6]
        np.testing.assert_allclose(slope, expected_slope, atol=1e-6)


def create_point_cloud_in_plane_and_neighborhood(nvect=None):
    if nvect is None:
        nvect = np.random.randn(3)

    nvect /= np.linalg.norm(nvect)
    points = _generate_random_points_in_plane(nvect, dparam=0, npts=100)
    neighborhood, pc = create_point_cloud_and_neighborhoods(points)
    return neighborhood, pc


def create_point_cloud_and_neighborhoods(point):
    pc = {keys.point: {'x': {'type': 'double', 'data': point[:, 0]},
                       'y': {'type': 'double', 'data': point[:, 1]},
                       'z': {'type': 'double', 'data': point[:, 2]}}}
    neighborhood = [[3, 4, 5, 6, 7], [1, 2, 7, 8, 9], [1, 7, 8, 9], [1, 2, 7, 9], [2, 7, 8, 9]]
    return neighborhood, pc


def _generate_random_points_in_plane(nvect, dparam, npts, eps=0.0):
    """
    Generate a series of point all belonging to a plane.

    :param nvect: normal vector of the plane
    :param dparam: zero point value of the plane
    :param npts: number of points
    :param eps: std of the gaussian noise added to the z values of the planes
    :return: x,y,z coordinate of the points
    """
    np.random.seed(12345)
    a, b, c = nvect / np.linalg.norm(nvect)
    x, y = np.random.rand(npts), np.random.rand(npts)
    z = (dparam - a * x - b * y) / c
    if eps > 0:
        z += np.random.normal(loc=0., scale=eps, size=npts)
    return np.column_stack((x, y, z))


class EigenValueSerial(FeatureExtractor):
    """Old serial implementation. Used to test the current (vectorized) implementation against."""

    @classmethod
    def requires(cls):
        return []

    @classmethod
    def provides(cls):
        return ["eigenv_1", "eigenv_2", "eigenv_3"]

    def extract(self, sourcepc, neighborhood, targetpc, targetindex, volume):
        nbptsX, nbptsY, nbptsZ = utils.get_point(sourcepc, neighborhood)
        matrix = np.column_stack((nbptsX, nbptsY, nbptsZ))

        try:
            eigenvals, eigenvecs = self._structure_tensor(matrix)
        except ValueError as err:
            if str(err) == 'Not enough points to compute eigenvalues/vectors.':
                return [0, 0, 0]
            else:
                raise

        return [eigenvals[0], eigenvals[1], eigenvals[2]]

    @staticmethod
    def _structure_tensor(points):
        """
        Computes the structure tensor of points by computing the eigenvalues
        and eigenvectors of the covariance matrix of a point cloud.
        Parameters
        ----------
        points : (Mx3) array
            X, Y and Z coordinates of points.
        Returns
        -------
        eigenvalues : (1x3) array
            The eigenvalues corresponding to the eigenvectors of the covariance
            matrix.
        eigenvectors : (3,3) array
            The eigenvectors of the covariance matrix.
        """
        if points.shape[0] > 3:
            cov_mat = np.cov(points, rowvar=False)
            eigenvalues, eigenvectors = np.linalg.eig(cov_mat)
            order = np.argsort(-eigenvalues)
            eigenvalues = eigenvalues[order]
            eigenvectors = eigenvectors[:, order]
            return eigenvalues, eigenvectors
        else:
            raise ValueError('Not enough points to compute eigenvalues/vectors.')



from scipy.spatial import cKDTree
import numpy as np
import weakref
from laserchicken import keys

kd_tree_cache = None


def get_kdtree_for_pc(pc):
    """
    Creates a kdtree of the point cloud based on its x and y attributes.
    :param pc: point cloud
    :return: kdtree object
    """
    index = -1
    xref = weakref.ref(pc[keys.point]["x"]["data"])
    yref = weakref.ref(pc[keys.point]["y"]["data"])
    for i in range(len(kd_tree_cache[0])):
        if xref is kd_tree_cache[0][i] and yref is kd_tree_cache[1][i]:
            index = i
    if index < 0:
        # TODO: Check if kd-tree is serialized to file...
        # If not, build it:
        kd_tree_cache[0].append(xref)
        kd_tree_cache[1].append(yref)
        kd_tree_cache[2].append(_build_kdtree(pc))
        index = len(kd_tree_cache[0]) - 1
    return kd_tree_cache[2][index]


def _build_kdtree(pc):
    points = np.column_stack((pc[keys.point]["x"].get("data", []), pc[keys.point]["y"].get("data", [])))
    return cKDTree(points, compact_nodes=False, balanced_tree=False)


def initialize_cache():
    global kd_tree_cache
    kd_tree_cache = ([], [], [])


initialize_cache()



from laserchicken.compute_neighbors import compute_neighborhoods
from laserchicken import keys
from laserchicken.feature_extractor.range_feature_extractor import RangeFeatureExtractor as range_extractor
from laserchicken.keys import normalized_height
import numpy as np

from laserchicken.test_tools import create_point_cloud
from laserchicken.utils import add_metadata
from laserchicken.volume_specification import Cell


def normalize(point_cloud, cell_size=None):
    z = point_cloud[keys.point]['z']['data']
    point_cloud[keys.point][normalized_height] = {"type": 'float64', "data": np.array(z)}
    if cell_size is None:
        n_points = point_cloud[keys.point][normalized_height]['data'].size
        min_z = _calculate_min_z(range(n_points), point_cloud)
        point_cloud[keys.point][normalized_height]['data'] = z - min_z
    else:
        targets = _create_spanning_grid(point_cloud, cell_size)

        neighborhoods = compute_neighborhoods(point_cloud, targets, Cell(cell_size), sample_size=None)
        for neighborhood in neighborhoods:
            min_z = _calculate_min_z(neighborhood, point_cloud)
            point_cloud[keys.point][normalized_height]['data'][neighborhood] = z[neighborhood] - min_z
    import sys
    module = sys.modules[__name__]
    add_metadata(point_cloud, module, {'cell_size':cell_size})
    return point_cloud


def _calculate_min_z(neighborhood, point_cloud):
    _, min_z, _ = range_extractor().extract(point_cloud, [neighborhood], None, None, None)
    return min_z[0]


def _create_spanning_grid(point_cloud, cell_size):
    x = point_cloud[keys.point]['x']['data']
    y = point_cloud[keys.point]['y']['data']
    min_x = np.min(x)
    max_x = np.max(x)
    min_y = np.min(y)
    max_y = np.max(y)

    cell_x_lengths, n_grid_points = _count_steps_and_points(cell_size, max_x, max_y, min_x, min_y)

    xs = [min_x + cell_size * (0.5 + (i % cell_x_lengths)) for i in range(n_grid_points)]
    ys = [min_y + cell_size * (0.5 + np.floor(i / cell_x_lengths)) for i in range(n_grid_points)]
    zs = np.zeros_like(xs)
    return create_point_cloud(xs, ys, zs)


def _count_steps_and_points(cell_size, max_x, max_y, min_x, min_y):
    cell_x_lengths = _count_steps(min_x, max_x, cell_size)
    cell_y_lengths = _count_steps(min_y, max_y, cell_size)
    n_grid_points = cell_x_lengths * cell_y_lengths
    return cell_x_lengths, n_grid_points


def _count_steps(min_x, max_x, cell_size):
    """Count the number of steps in a grid in a single direction."""
    return max(int(np.ceil((max_x - min_x) / float(cell_size))), 1)



"""Tests for the feature_extractor module."""
from .feature_test1 import Test1FeatureExtractor
from .feature_test23 import Test2FeatureExtractor, Test3FeatureExtractor, TestVectorizedFeatureExtractor
from .feature_test_broken import TestBrokenFeatureExtractor



"""Broken feature extractor."""
from laserchicken.feature_extractor.base_feature_extractor import FeatureExtractor


class TestBrokenFeatureExtractor(FeatureExtractor):
    """Feature extractor that fails to add the feature in promises to provide to target."""

    @classmethod
    def requires(cls):
        return []

    @classmethod
    def provides(cls):
        return ['test_broken']

    def extract(self, sourcepc, neighborhood, targetpc, targetindex, volume):
        pass



import os
import unittest

from distutils.version import StrictVersion

import laserchicken

class TestVersion(unittest.TestCase):
    version_file_path = None

    def test_version_file_exists(self):
        self.assertTrue(os.path.isfile(self.version_file_path))

    def test_version_file_is_readable(self):
        with open(self.version_file_path) as f:
            readable = f.readable()
        self.assertTrue(readable)

    def test_version_file_is_not_empty(self):
        self.assertTrue(self._read_version(self.version_file_path))

    def test_version_is_correct(self):
        version_read = StrictVersion(self._read_version(self.version_file_path))
        verson_import = laserchicken.__version__
        self.assertEqual(version_read, verson_import)

    @staticmethod
    def _read_version(file_path):
        with open(file_path) as f:
            lines = f.read()
        return lines.strip()

    def setUp(self):
        cwd = os.path.dirname(__file__)
        self.version_file_path = os.path.join(cwd, '_version.txt')


import unittest


class Test(unittest.TestCase):
    def test_config_test(self):
        """ This is the first test of the project and should pass if test runner is configured correctly. """
        pass



import sys
from laserchicken.io import get_io_handler
from laserchicken.utils import add_metadata


def load(path, format=None, *args, **kwargs):
    """
    Read point cloud from a file.

    :param path:
    :param format:
    :param args: optional non-keyword arguments to be passed to the format-specific writer
    :param kwargs: optional keyword arguments to be passed to the format-specific writer
    :return: point cloud data
    """
    reader = get_io_handler(path, mode='r', format=format)
    point_cloud = reader.read(*args, **kwargs)
    add_metadata(point_cloud, sys.modules[__name__], {'path': path, 'args': args, **kwargs})
    return point_cloud
